#!/usr/bin/env python3
"""Tests for operations.timeseries."""

import numpy as np
from numpy.random import rand
from pytest import mark, raises

from nigsp.operations import timeseries
from nigsp.utils import prepare_ndim_iteration


# ### Unit tests
def test_normalise_ts():
    ts = rand(3, 20)
    z = (ts - ts.mean(axis=1)[:, np.newaxis, ...]) / ts.std(axis=1, ddof=1)[
        :, np.newaxis, ...
    ]
    z[np.isnan(z)] = 0

    tsz = timeseries.normalise_ts(ts)

    assert (tsz == z).all()


def test_graph_fourier_transform():
    ts = rand(3, 6)
    ev = rand(3, 3)
    pr = ev.conj().T @ ts

    proj = timeseries.graph_fourier_transform(ts, ev)

    assert (proj == pr).all()

    ts = rand(3, 6, 2, 2)
    tsr, pr = prepare_ndim_iteration(ts, 2)

    for i in range(tsr.shape[-1]):
        pr[:, :, i] = ev.conj().T @ np.squeeze(tsr[:, :, i])
    pr = pr.reshape(ts.shape)
    prm = pr.mean(axis=1)
    engm = (pr**2).mean(axis=1)

    proj = timeseries.graph_fourier_transform(ts, ev, mean=True)
    eng = timeseries.graph_fourier_transform(ts, ev, energy=True, mean=True)

    assert (proj == prm).all()
    assert (eng == engm).all()


def test_median_cutoff_frequency_idx():
    energy = np.array([[0, 3], [1, 2], [0, 1], [0.5, 0.5]])
    freq_t = 2

    freq_idx = timeseries.median_cutoff_frequency_idx(energy)

    assert freq_t == freq_idx


def test_graph_filter():
    ts = rand(4, 10)
    ev = rand(4, 4)

    keys = ["hi", "ho"]
    freq_idx = 2

    ev_s = dict.fromkeys(keys)
    ts_s = dict.fromkeys(keys)

    ev_s["hi"] = np.append(
        ev[:, :freq_idx], np.zeros_like(ev[:, freq_idx:], dtype="float32"), axis=-1
    )
    ev_s["ho"] = np.append(
        np.zeros_like(ev[:, :freq_idx], dtype="float32"), ev[:, freq_idx:], axis=-1
    )

    fcf = timeseries.graph_fourier_transform(ts, ev)
    ts_s["ho"] = timeseries.graph_fourier_transform(fcf, ev_s["ho"].T)
    ts_s["hi"] = timeseries.graph_fourier_transform(fcf, ev_s["hi"].T)

    evec_split, ts_split = timeseries.graph_filter(ts, ev, freq_idx, keys)

    assert all(item in keys for item in list(evec_split.keys()))
    assert all(item in keys for item in list(ts_split.keys()))
    for k in keys:
        assert (ev_s[k] == evec_split[k]).all()
        assert (ts_s[k] == ts_split[k]).all()


def test_functional_connectivity():
    assert timeseries.functional_connectivity(rand(6)) == 1

    ts = rand(3, 6, 2, 2)

    tst, _ = prepare_ndim_iteration(ts, 2)
    fc = np.empty(([tst.shape[0]] * 2 + [tst.shape[-1]]), dtype="float32")
    for i in range(tst.shape[-1]):
        fc[:, :, i] = np.corrcoef(tst[..., i])

    ns = (ts.shape[0],) * 2 + ts.shape[2:]
    fc = fc.reshape(ns).mean(axis=2).squeeze()

    fc_t = timeseries.functional_connectivity(ts, mean=True)

    assert (fc == fc_t).all()

    tsd = {"hi": rand(3, 6), "lo": rand(3, 6)}

    fcd = timeseries.functional_connectivity(tsd)

    assert all(item in list(tsd.keys()) for item in list(fcd.keys()))

    for k in fcd.keys():
        assert (fcd[k] == np.corrcoef(tsd[k])).all()


# ### Break tests
def test_break_median_cutoff_frequency_idx():
    with raises(NotImplementedError) as errorinfo:
        timeseries.median_cutoff_frequency_idx(rand(2, 3, 4))
    assert "have 3 dimensions" in str(errorinfo.value)


@mark.parametrize("freq", [(0), (2), (4)])
def test_break_graph_filter(freq):
    with raises(IndexError) as errorinfo:
        timeseries.graph_filter(rand(2, 3), rand(2, 2), freq)
    assert f"index {freq} is not valid" in str(errorinfo.value)



#!/usr/bin/env python3
"""
Module to plot graphs.

All `viz` functions require the extra modules `matplotlib` and `nibabel`,
installable using the flag:
```shell
$ pip3 install nigsp[viz]
```

Attributes
----------
FIGSIZE : tuple
    Figure size
LGR
    Logger
SET_DPI : int
    DPI of the figure
"""

import logging

import numpy as np

from nigsp.operations.timeseries import spc_ts

LGR = logging.getLogger(__name__)
SET_DPI = 100
FIGSIZE = (18, 10)
FIGSIZE_SQUARE = (6, 5)
FIGSIZE_LONG = (10, 5)


def plot_connectivity(mtx, filename=None, closeplot=False):
    """
    Create a connectivity matrix plot.

    If mtx has 3 dimensions, average first along the last axis.

    Parameters
    ----------
    mtx : numpy.ndarray
        A (square) array with connectivity information inside.
    filename : None, str, or os.PathLike, optional
        The path to save the plot on disk.
    closeplot : bool, optional
        Whether to close plots after saving or not. Mainly used for debug
        or use with live python/ipython instances.

    Returns
    -------
    0
        If there are no errors.

    Raises
    ------
    ImportError
        If matplotlib and/or nilearn are not installed.
    ValueError
        If mtx has more than 3 dimensions.

    Notes
    -----
    Requires `matplotlib` and `nilearn`
    """
    try:
        import matplotlib.pyplot as plt
        from nilearn.plotting import plot_matrix
    except ImportError:
        raise ImportError(
            "nilearn and matplotlib are required to plot node images. "
            "Please see install instructions."
        )

    mtx = mtx.squeeze()
    if mtx.ndim > 3:
        raise ValueError(
            "Cannot plot connectivity matrices for matrix of dimensions > 3."
        )
    elif mtx.ndim == 3:
        LGR.warning("Since matrix is 3D, averaging across last dimension.")
        mtx = mtx.mean(axis=-1)

    if mtx.shape[0] != mtx.shape[1]:
        LGR.warning("Given matrix is not a square matrix!")

    LGR.info("Creating connectivity plot.")
    plt.figure(figsize=FIGSIZE_SQUARE)
    plot_matrix(mtx)

    if filename is not None:
        plt.savefig(filename, dpi=SET_DPI)
        closeplot = True

    if closeplot:
        plt.close()

    return 0


def plot_greyplot(timeseries, filename=None, title=None, spc=True, closeplot=False):
    """
    Create a greyplot (a.k.a. carpet plot a.k.a. timeseries plot).

    If timeseries has 3 dimensions, average first along the last axis.

    Parameters
    ----------
    timeseries : numpy.ndarray
        An array representing a timeseries. Time has to be encoded in the
        second dimension.
    filename : None, str, or os.PathLike, optional
        The path to save the plot on disk.
    title : None or str, optional
        Add a title to the graph
    normalise : bool, optional
        If true, spc the signal before plotting.
    closeplot : bool, optional
        Whether to close plots after saving or not. Mainly used for debug
        or use with live python/ipython instances.

    Returns
    -------
    0
        If there are no errors.

    Raises
    ------
    ImportError
        If matplotlib is not installed.
    ValueError
        If timeseries has more than 3 dimensions.

    Notes
    -----
    Requires `matplotlib`
    """
    try:
        import matplotlib.pyplot as plt
    except ImportError:
        raise ImportError(
            "matplotlib is required to plot node images. "
            "Please see install instructions."
        )

    timeseries = timeseries.squeeze()
    if timeseries.ndim > 3:
        raise ValueError("Cannot plot greyplots for timeseries of dimensions > 3.")
    elif timeseries.ndim == 3:
        LGR.warning("Since timeseries is 3D, averaging across last dimension.")
        timeseries = timeseries.mean(axis=-1)

    if spc:
        LGR.info("Expressing timeseries in signal percentage change")
        timeseries = spc_ts(timeseries)

    LGR.info("Creating greyplot.")
    plt.figure(figsize=FIGSIZE_LONG)
    if title is not None:
        plt.title(title)
    vmax = np.percentile(timeseries, 99)
    vmin = np.percentile(timeseries, 1)
    plt.imshow(timeseries, cmap="gray", vmin=vmin, vmax=vmax)
    plt.tight_layout()

    if filename is not None:
        plt.savefig(filename, dpi=SET_DPI)
        closeplot = True

    if closeplot:
        plt.close()

    return 0


def plot_nodes(ns, atlas, filename=None, thr=None, closeplot=False):
    """
    Create a marker plot in the MNI space.

    If ns has 2 dimensions, average first along last dimension.

    Parameters
    ----------
    ns : numpy.ndarray
        A 1- or 2- D array that contains the value of the nodes.
    atlas : str, os.PathLike, 3D Nifti1Image, or numpy.ndarray
        The 3d nifti image of an atlas, a string or path to its position,
        or a list of coordinates of the center of mass of parcels.
    filename : None, str, or os.PathLike, optional
        The path to save the plot on disk.
    thr : float or None, optional
        The threshold to use in plotting the nodes.
    closeplot : bool, optional
        Whether to close plots after saving or not. Mainly used for debug
        or use with live python/ipython instances.

    Returns
    -------
    0
        If there are no errors.

    Raises
    ------
    ImportError
        If matplotlib and/or nilearn are not installed.
    ValueError
        If ns has more than 2 dimensions.
        If coordinates can't be extracted from atlas.

    Notes
    -----
    Requires `matplotlib` and `nilearn`
    """
    try:
        import matplotlib.pyplot as plt
        from nilearn.plotting import find_parcellation_cut_coords, plot_markers
    except ImportError:
        raise ImportError(
            "nilearn and matplotlib are required to plot node images. "
            "Please see install instructions."
        )
    # First check that ns is a valid source of data.
    ns = ns.squeeze()
    if ns.ndim > 2:
        raise ValueError("Cannot plot node values for matrix of dimensions > 2.")
    elif ns.ndim == 2:
        LGR.warning("Given matrix has 2 dimensions, averaging across last dimension.")
        ns = ns.mean(axis=-1)

    # Then treat atlas
    if type(atlas) is np.ndarray:
        if atlas.ndim > 2 or atlas.shape[1] != 3:
            raise NotImplementedError(
                "Only atlases in nifti format or list of coordinates are supported."
            )
        else:
            coord = atlas
    else:
        coord = find_parcellation_cut_coords(atlas)

    if ns.shape[0] != coord.shape[0]:
        raise ValueError("Node array and coordinates array have different length.")

    LGR.info("Creating markerplot.")
    plt.figure(figsize=FIGSIZE)
    plot_markers(ns, coord, node_threshold=thr)

    if filename is not None:
        plt.savefig(filename, dpi=SET_DPI)
        closeplot = True

    if closeplot:
        plt.close()

    return 0


def plot_edges(mtx, atlas, filename=None, thr=None, closeplot=False):
    """
    Create a connectivity plot in the MNI space.

    If mtx has 2 dimensions, average first along last dimension.

    Parameters
    ----------
    mtx : numpy.ndarray
        A 2- or 3- D array that contains the value of the nodes.
    atlas : str, os.PathLike, 3D Nifti1Image, or numpy.ndarray
        The 3d nifti image of an atlas, a string or path to its position,
        or a list of coordinates of the center of mass of parcels.
    filename : None, str, or os.PathLike, optional
        The path to save the plot on disk.
    thr : float, str or None, optional
        The threshold to use in plotting the nodes.
        If `str`, needs to express a percentage.
    closeplot : bool, optional
        Whether to close plots after saving or not. Mainly used for debug
        or use with live python/ipython instances.

    Returns
    -------
    0
        If there are no errors.

    Raises
    ------
    ImportError
        If matplotlib and/or nilearn are not installed.
    ValueError
        If mtx has more than 3 dimensions.
        If coordinates can't be extracted from atlas.

    Notes
    -----
    Requires `matplotlib` and `nilearn`
    """
    try:
        import matplotlib.pyplot as plt
        from nilearn.plotting import cm, find_parcellation_cut_coords, plot_connectome
    except ImportError:
        raise ImportError(
            "nilearn and matplotlib are required to plot node images. "
            "Please see install instructions."
        )
    # First check that mtx is a valid source of data.
    mtx = mtx.squeeze()
    if mtx.ndim > 3:
        raise ValueError("Cannot plot node values for matrix of dimensions > 3.")
    elif mtx.ndim == 3:
        LGR.warning("Given matrix has 3 dimensions, averaging across last dimension.")
        mtx = mtx.mean(axis=-1)

    # Then treat atlas
    if type(atlas) is np.ndarray:
        if atlas.ndim > 2 or atlas.shape[1] != 3:
            raise NotImplementedError(
                "Only atlases in nifti format or list of coordinates are supported."
            )
        else:
            coord = atlas
    else:
        coord = find_parcellation_cut_coords(atlas)

    if mtx.shape[0] != coord.shape[0]:
        raise ValueError("Matrix axis and coordinates array have different length.")

    LGR.info("Creating connectome-like plot.")
    plt.figure(figsize=FIGSIZE)

    pc_args = {
        "adjacency_matrix": mtx,
        "node_coords": coord,
        "node_color": "black",
        "node_size": 5,
        "edge_threshold": thr,
        "colorbar": True,
    }

    if mtx.min() >= 0:
        pc_args["edge_vmin"] = 0
        pc_args["edge_vmax"] = mtx.max()
        pc_args["edge_cmap"] = cm.red_transparent_full_alpha_range

    plot_connectome(**pc_args)

    if filename is not None:
        plt.savefig(filename, dpi=SET_DPI)
        closeplot = True

    if closeplot:
        plt.close()

    return 0


"""
Copyright 2022, Stefano Moia.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""



#!/usr/bin/env python3
"""Tests for objects."""

import numpy as np
from numpy.random import rand
from pytest import raises

from nigsp import operations
from nigsp.objects import SCGraph


# ### Unit tests
def test_SCGraph():
    """Test SCGraph, properties, and methods."""
    # # Initialise object
    # Initialise content
    mtx = rand(4, 4)
    timeseries = rand(4, 6)
    atlas = rand(4, 3)
    filename = "Laudna.nii.gz"
    ts_split = {"high": rand(4, 6), "low": rand(4, 6)}
    # Initialise eigenvec to check zerocross
    def _bonnet(d, x):
        if d == 0:
            return np.ones_like(x)
        elif d == 1:
            return x
        else:
            return (
                (2 * d - 1) * x * _bonnet(d - 1, x) - (d - 1) * _bonnet(d - 2, x)
            ) / d

    x = np.linspace(-1, 1, 4)
    eigenvec = np.empty([4, 4], dtype="float32")
    for n in range(4):
        eigenvec[:, n] = _bonnet(n, x)
    zx = np.linspace(0, 3, 4)

    # Initialise SCGraph proper
    scgraph = SCGraph(
        mtx,
        timeseries,
        atlas=atlas,
        filename=filename,
        eigenvec=eigenvec,
        ts_split=ts_split,
        index=2,
    )

    # # Assert properties
    assert scgraph.nnodes == mtx.shape[1]
    assert scgraph.ntimepoints == timeseries.shape[1]
    assert (scgraph.zerocross == zx).all()
    assert all(item in scgraph.split_keys for item in list(ts_split.keys()))

    # # Test methods
    # Only test split_graph, create_surrogates, compute_fc
    # Other methods are somewhat one-liner wrap-around of tested functions.

    # Test split_graph and split_graph index priority
    scgraph.split_graph()
    evs, tss = operations.graph_filter(timeseries, eigenvec, 2)
    # Update ts_split to be tss
    ts_split = tss

    assert all(item in list(evs.keys()) for item in list(scgraph.evec_split.keys()))
    assert all(item in list(tss.keys()) for item in list(scgraph.ts_split.keys()))
    assert (evs["low"] == scgraph.evec_split["low"]).all()
    assert (tss["low"] == scgraph.ts_split["low"]).all()

    scgraph.evec_split = {}
    scgraph.index = "median"
    scgraph.split_graph(index=2)

    assert (evs["low"] == scgraph.evec_split["low"]).all()

    # Test create_surrogates
    scgraph.create_surrogates(sc_type="informed", n_surr=1, seed=6)
    i_surr = operations.sc_informed(timeseries, eigenvec, n_surr=1, seed=6)
    assert (scgraph.surr == i_surr).all()

    scgraph.lapl_mtx = mtx
    scgraph.create_surrogates(sc_type="uninformed", n_surr=1, seed=6)
    u_surr = operations.sc_uninformed(timeseries, lapl_mtx=mtx, n_surr=1, seed=6)
    assert (scgraph.surr == u_surr).all()

    # Test compute_fc
    scgraph.compute_fc()
    fc = operations.functional_connectivity(timeseries)
    fc_low = operations.functional_connectivity(ts_split["low"])

    assert (fc == scgraph.fc).all()
    assert all(item in ["high", "low"] for item in list(scgraph.fc_split.keys()))
    assert (fc_low == scgraph.fc_split["low"]).all()


# ### Break tests
def test_break_SCGraph():
    """Break SCGraph and its methods."""
    with raises(ValueError) as errorinfo:
        SCGraph(rand(3, 4), rand(4, 6))
    assert "square matrix" in str(errorinfo.value)

    with raises(ValueError) as errorinfo:
        SCGraph(rand(4, 4), rand(3, 6))
    assert "number of parcels and nodes" in str(errorinfo.value)

    with raises(ValueError) as errorinfo:
        SCGraph(rand(4, 4), rand(4, 6, 4, 5))
    assert "more than 3 dimensions" in str(errorinfo.value)

    scgraph = SCGraph(rand(4, 4), rand(4, 6), index="Chet")
    with raises(ValueError) as errorinfo:
        scgraph.split_graph()
    assert "Unknown option Chet" in str(errorinfo.value)

    scgraph = SCGraph(rand(4, 4), rand(4, 6))
    with raises(ValueError) as errorinfo:
        scgraph.create_surrogates(sc_type="Fearne")
    assert "Unknown option Fearne" in str(errorinfo.value)



#!/usr/bin/env python3
"""Integration test."""
import shutil
from os.path import isdir, isfile, join

import numpy as np

from nigsp.workflow import _main


# ### Integration tests
def test_integration(timeseries, sc_mtx, atlas, mean_fc, sdi, testdir):
    """Integration test for FULL workflow."""
    testdir = join(testdir, "testdir")
    mean_fc_mat = np.genfromtxt(mean_fc)
    # Compared to original SDI, now we log2 it).
    sdi_mat = np.log2(np.genfromtxt(sdi))

    # Run workflow
    # fmt: off
    _main(
        [
            "-f", timeseries,
            "-s", sc_mtx,
            "-a", atlas,
            "-odir", testdir,
            "-ofile", "testfile.tsv",
            "-sci",
            "-n", "4",
            "-seed", "42",
        ]
    )
    # fmt: on
    # Check that files were created
    assert isdir(testdir)
    assert isdir(join(testdir, "logs"))
    assert isdir(join(testdir, "testfile_timeseries_low"))
    assert isdir(join(testdir, "testfile_timeseries_high"))
    assert isfile(join(testdir, "testfile_timeseries_low", "000.tsv"))
    assert isfile(join(testdir, "testfile_timeseries_high", "000.tsv"))
    assert isfile(join(testdir, "testfile_fc.tsv"))
    assert isfile(join(testdir, "testfile_fc_low.tsv"))
    assert isfile(join(testdir, "testfile_fc_high.tsv"))
    assert isfile(join(testdir, "testfile_eigenval.tsv"))
    assert isfile(join(testdir, "testfile_eigenvec.tsv"))
    assert isfile(join(testdir, "testfile_eigenvec_low.tsv"))
    assert isfile(join(testdir, "testfile_eigenvec_high.tsv"))
    assert isfile(join(testdir, "testfile_sdi.tsv"))
    assert isfile(join(testdir, "testfile_mkd_sdi.tsv"))
    assert isfile(join(testdir, "testfile_laplacian.png"))
    assert isfile(join(testdir, "testfile_sc.png"))
    assert isfile(join(testdir, "testfile_fc.png"))
    assert isfile(join(testdir, "testfile_fc_low.png"))
    assert isfile(join(testdir, "testfile_fc_high.png"))
    assert isfile(join(testdir, "testfile_greyplot.png"))
    assert isfile(join(testdir, "testfile_greyplot_low.png"))
    assert isfile(join(testdir, "testfile_greyplot_high.png"))
    assert isfile(join(testdir, "testfile_sdi.png"))
    assert isfile(join(testdir, "testfile_mkd_sdi.png"))

    sdi_pyt = np.genfromtxt(join(testdir, "testfile_sdi.tsv"))
    sdi_mkd = np.genfromtxt(join(testdir, "testfile_mkd_sdi.tsv"))

    mean_fc_pyt = np.genfromtxt(join(testdir, "testfile_fc.tsv"))
    # Check that each cell in the result is comparable to matlab's.
    # There's a bunch of rounding due to np.round and numerical difference between matlab and python
    assert abs(mean_fc_pyt.round(6) - mean_fc_mat.round(6)).max().round(6) <= 0.000001
    assert abs(sdi_pyt.round(5) - sdi_mat.round(5)).max().round(5) <= 0.00001
    assert (sdi_pyt != sdi_mkd).any()

    # Clean up!
    shutil.rmtree(testdir)



#!/usr/bin/env python3
"""Tests for io."""

import sys
from os import remove
from os.path import isfile

import nibabel
import pymatreader
import scipy
from numpy import asarray, empty, genfromtxt, savetxt
from numpy.random import rand
from pytest import mark, raises

from nigsp import io

# ### Unit tests


def test_check_ext():
    """Text check_ext."""
    all_ext = io.EXT_1D + io.EXT_MAT + io.EXT_NIFTI
    for ext_in in all_ext:
        fname_in = f"shiny{ext_in}"

        for fname in [fname_in, fname_in.upper()]:
            has_ext, fname_out = io.check_ext(all_ext, fname)

            assert has_ext is True
            assert fname_out == fname

            has_ext, fname_out, ext_out = io.check_ext(all_ext, fname, remove=True)

            assert has_ext is True
            assert fname_out == fname[: -len(ext_in)]
            assert ext_out == ext_in

    # #!# Missing scan option

    fname_in = "shiny.mal"
    for fname in [fname_in, fname_in.upper()]:
        has_ext, fname_out = io.check_ext(".inara", fname)

        assert has_ext is False
        assert fname_out == fname

        has_ext, fname_out, ext_out = io.check_ext(".inara", fname, remove=True)

        assert has_ext is False
        assert fname_out == fname
        assert ext_out == ""


@mark.parametrize("data", [(rand(3, 4)), (rand(3, 4, 1)), (rand(3, 1, 4))])
def test_check_nifti_dim(data):
    """Test check_nifti_dim."""
    data_out = io.check_nifti_dim("kaylee", data, dim=2)

    assert (data_out == data.squeeze()).all()
    assert data_out.ndim == 2


@mark.parametrize("data", [(rand(3, 4)), (rand(3, 4, 1)), (rand(3, 1, 4))])
def test_check_mtx_dim(data):
    """Test check_mtx_dim."""
    data_out = io.check_mtx_dim("wash", data)

    assert (data_out == data.squeeze()).all()
    assert data_out.ndim == 2


@mark.parametrize("data, shape", [(rand(3), "rectangle"), (rand(3, 3), "square")])
def test_check_mtx_dim_cases(data, shape):
    """Test check_mtx_dim for specific shapes."""
    data_out = io.check_mtx_dim("wash", data, shape=shape)

    assert (data_out.squeeze() == data).all()
    assert data_out.ndim == 2


def test_load_nifti_get_mask(atlas):
    """Test load_nifti_get_mask."""
    img = nibabel.load(atlas)
    data = img.get_fdata()
    mask = data.any(axis=-1).squeeze()

    d, m, i = io.load_nifti_get_mask(atlas, ndim=3)

    assert (data == d).all()
    assert (mask == m).all()
    assert (img.header["dim"] == i.header["dim"]).all()

    remove(atlas)


def test_load_txt():
    """Test load_txt."""
    a = rand(5)
    n = "zoe"
    savetxt(n, a)
    b = io.load_txt(n)

    assert (a == b).all()

    remove(n)


def test_load_mat():
    """Test load_mat."""
    a = rand(5)
    b = "likealeaf"
    n = "inthewind"

    scipy.io.savemat(n, {"data": a, "other": b})

    c = io.load_mat(n)

    assert (a == c).all()
    remove(n)


def test_export_nifti(atlas):
    """Test export_nifti."""
    img = nibabel.load(atlas)
    shape = img.get_fdata().shape
    io.export_nifti(empty(shape), img, "book")
    assert isfile("book.nii.gz")
    remove("book.nii.gz")
    remove(atlas)


@mark.parametrize(
    "ext_in, ext_out",
    [
        (".1D", ".1D"),
        (".csv", ".csv"),
        (".tsv", ".tsv"),
        ("", ".csv"),
        (".mat", ".mat"),
    ],
)
def test_export_mtx(ext_in, ext_out):
    """Test export_mtx."""
    data = asarray([[1, 1, 2], [3, 5, 8]])
    io.export_mtx(data, "serenity", ext=ext_in)
    assert isfile(f"serenity{ext_out}")

    if ext_out in [".csv"]:
        data_in = genfromtxt(f"serenity{ext_out}", delimiter=",")
    if ext_out in [".tsv", ".1D"]:
        data_in = genfromtxt(f"serenity{ext_out}")

    if ext_out in [".mat"]:
        data_in = pymatreader.read_mat(f"serenity{ext_out}")
        data_in = data_in["data"]

    assert (data_in == data).all()
    # remove data
    remove(f"serenity{ext_out}")


# ### Break tests
@mark.parametrize("data", [(rand(3)), (rand(3, 4, 2))])
def test_break_check_nifti_dim(data):
    """Break check_nifti_dim."""
    with raises(ValueError) as errorinfo:
        io.check_nifti_dim("jayne", data, dim=2)
    assert f"jayne is {data.ndim}D" in str(errorinfo.value)


def test_break_check_mtx_dim():
    """Break check_mtx_dim."""
    with raises(ValueError) as errorinfo:
        io.check_mtx_dim("river", empty(0))
    assert "river is empty" in str(errorinfo.value)

    with raises(NotImplementedError) as errorinfo:
        io.check_mtx_dim("river", rand(3, 4, 5, 6))
    assert "3D are supported" in str(errorinfo.value)

    with raises(ValueError) as errorinfo:
        io.check_mtx_dim("river", rand(3, 4), shape="square")
    assert "river matrix has shape (3, 4)" in str(errorinfo.value)


def test_break_load_nifti_get_mask():
    """Break load_nifti_get_mask."""
    sys.modules["nibabel"] = None
    with raises(ImportError) as errorinfo:
        io.load_nifti_get_mask("reavers")
    assert "is required" in str(errorinfo.value)
    sys.modules["nibabel"] = nibabel


def test_break_load_mat():
    """Break load_mat."""
    sys.modules["pymatreader"] = None
    with raises(ImportError) as errorinfo:
        io.load_mat("simon")
    assert "is required" in str(errorinfo.value)
    sys.modules["pymatreader"] = pymatreader

    a = "heart"
    n = "ofgold"
    scipy.io.savemat(n, {"data": a})

    with raises(EOFError) as errorinfo:
        io.load_mat(n)
    assert f"{n} does not seem" in str(errorinfo.value)
    remove(n)


def test_break_load_xls():
    """Break load_xls."""
    with raises(NotImplementedError) as errorinfo:
        io.load_xls("firefly")
    assert "loading is not" in str(errorinfo.value)


def test_break_export_nifti():
    """Break export_nifti."""
    sys.modules["nibabel"] = None
    with raises(ImportError) as errorinfo:
        io.export_nifti(rand(3), None, "reavers")
    assert "is required" in str(errorinfo.value)
    sys.modules["nibabel"] = nibabel


def test_break_export_mtx():
    """Break export_mtx."""
    with raises(NotImplementedError) as errorinfo:
        io.export_mtx(rand(3, 4), "lostinthewoods", ext=".xls")
    assert "output is not" in str(errorinfo.value)

    sys.modules["scipy"] = None
    with raises(ImportError) as errorinfo:
        io.export_mtx(rand(3, 4), "lostinthewoods", ext=".mat")
    assert "is required" in str(errorinfo.value)
    sys.modules["scipy"] = scipy



#!/usr/bin/env python3
"""Tests for operations.laplacian."""
from copy import deepcopy as dc

import numpy as np
from pytest import mark, raises

from nigsp.operations import laplacian


# ### Unit tests
def test_compute_laplacian():
    def glap(mtx):
        deg = mtx.sum(axis=1)
        adj = dc(mtx)
        adj[np.diag_indices(mtx.shape[0])] = 0

        L = np.diag(deg) - adj
        return L, deg

    mtx = np.random.rand(4, 4)
    mtx = (mtx + mtx.T) / 2

    L, deg = glap(mtx)
    lapl, degree = laplacian.compute_laplacian(mtx)

    assert (lapl == L).all()
    assert (degree == deg).all()

    mtx = mtx - mtx.mean()

    mtx_abs = abs(mtx)
    mtx_rem = dc(mtx)
    mtx_rem[mtx < 0] = 0
    mtx_res = (mtx - mtx.min()) / mtx.max()

    L, deg = glap(mtx_abs)
    lapl, degree = laplacian.compute_laplacian(mtx, negval="absolute")
    assert (lapl == L).all()
    assert (degree == deg).all()

    L, deg = glap(mtx_rem)
    lapl, degree = laplacian.compute_laplacian(mtx, negval="remove")
    assert (lapl == L).all()
    assert (degree == deg).all()

    L, deg = glap(mtx_res)
    lapl, degree = laplacian.compute_laplacian(mtx, negval="rescale")
    assert (lapl == L).all()
    assert (degree == deg).all()


@mark.xfail
def test_normalisation():

    L = np.random.rand(4, 4)
    L = (L + L.T) / 2
    d = np.random.rand(4)
    d[2] = 0

    lapl_symm = laplacian.normalisation(L, d, norm="symmetric")

    d = np.diag(d)
    lapl_rw = laplacian.normalisation(L, d, norm="random walk")

    d[2] = 1
    d_symm = np.diag(d ** (-1 / 2))
    d_rw = np.diag(d ** (-1))

    assert (lapl_symm == (d_symm @ L @ d_symm)).all()
    assert (lapl_rw == d_rw @ L).all()


def test_symmetric_normalised_laplacian():
    mtx = np.random.rand(4, 4)
    mtx = (mtx + mtx.T) / 2

    d = np.diag(mtx.sum(axis=-1) ** (-1 / 2))

    symm_norm = d @ mtx @ d

    lapl_in = np.eye(mtx.shape[0]) - symm_norm
    lapl_out = laplacian.symmetric_normalised_laplacian(mtx)

    assert (lapl_in == lapl_out).all()

    dg1 = np.diag(np.ones(mtx.shape[0]))
    lapl_in = np.eye(mtx.shape[0]) - (dg1 @ mtx @ dg1)
    lapl_out = laplacian.symmetric_normalised_laplacian(mtx, dg1)

    assert (lapl_in == lapl_out).all()

    dgv1 = np.ones(mtx.shape[0])
    lapl_out = laplacian.symmetric_normalised_laplacian(mtx, dgv1)

    assert (lapl_in == lapl_out).all()


def test_decomposition():
    mtx = np.random.rand(4, 4)
    mtx = (mtx + mtx.T) / 2

    eival, eivec = np.linalg.eig(mtx)

    idx = np.argsort(eival)
    eival = eival[idx]
    eivec = eivec[:, idx]

    eigenval, eigenvec = laplacian.decomposition(mtx)

    assert (eival == eigenval).all()
    assert (eivec == eigenvec).all()


def test_recomposition():
    mtx = np.random.rand(4, 4)
    mtx = (mtx + mtx.T) / 2

    eival, eivec = np.linalg.eig(mtx)

    idx = np.argsort(eival)
    eival = eival[idx]
    eivec = eivec[:, idx]

    mtx_rec = laplacian.recomposition(eival, eivec)

    assert abs(mtx - mtx_rec).sum().round(6) < 10**-6


# ### Break tests
def test_break_compute_laplacian():
    mtx = np.random.rand(4, 4)
    mtx = mtx - mtx.mean()
    with raises(NotImplementedError) as errorinfo:
        laplacian.compute_laplacian(mtx, negval="random")
    assert 'Behaviour "random" to deal' in str(errorinfo.value)


def test_break_normalisation():
    lapl = np.random.rand(4, 4, 4)
    d = np.random.rand(4)
    with raises(NotImplementedError) as errorinfo:
        laplacian.normalisation(lapl, d)
    assert "degree matrix is 1D" in str(errorinfo.value)

    lapl = np.random.rand(4, 4)
    d = np.diag(d)
    d[1, 2] = 1
    with raises(ValueError) as errorinfo:
        laplacian.normalisation(lapl, d)
    assert "not a diagonal matrix" in str(errorinfo.value)

    d = np.random.rand(6)
    with raises(ValueError) as errorinfo:
        laplacian.normalisation(lapl, d)
    assert "degree matrix has shape" in str(errorinfo.value)

    d = np.random.rand(4)
    with raises(NotImplementedError) as errorinfo:
        laplacian.normalisation(lapl, d, norm="echo")
    assert 'Normalisation type "echo"' in str(errorinfo.value)


def test_break_symmetric_normalised_laplacian():

    mtx = np.random.rand(4, 4)

    d = np.diag(np.ones(4), k=1)
    with raises(ValueError) as errorinfo:
        laplacian.symmetric_normalised_laplacian(mtx, d)
    assert "not a diagonal" in str(errorinfo.value)

    d = np.diag(np.ones(3))
    with raises(ValueError) as errorinfo:
        laplacian.symmetric_normalised_laplacian(mtx, d)
    assert "has shape" in str(errorinfo.value)


def test_break_recomposition():

    eivec = np.random.rand(4, 4, 2)
    eival = np.random.rand(4)

    with raises(NotImplementedError) as errorinfo:
        laplacian.recomposition(eival, eivec)
    assert "matrix dimensionality (3)" in str(errorinfo.value)

    eivec = np.random.rand(4, 4)
    # with raises(ValueError) as errorinfo:
    #     laplacian.recomposition(eival, eivec)
    # assert "Not enough dimensions" in str(errorinfo.value)

    eival = np.random.rand(4, 4)
    with raises(ValueError) as errorinfo:
        laplacian.recomposition(eival, eivec)
    assert "not a diagonal" in str(errorinfo.value)

    eival = np.random.rand(4, 4, 6)
    with raises(ValueError) as errorinfo:
        laplacian.recomposition(eival, eivec)
    assert "Too many dimensions" in str(errorinfo.value)



#!/usr/bin/env python3
"""
Operations on nifti files.

Attributes
----------
LGR
    Logger
"""

import logging

import numpy as np

LGR = logging.getLogger(__name__)


def vol_to_mat(data):
    """
    Reshape <3D in 1D or 4D into 2D.

    Parameters
    ----------
    data : numpy.ndarray
        The data to be transformed into 2D.

    Returns
    -------
    numpy.ndarray
        2D reshaped data.
    """
    LGR.info(f"Reshape {data.ndim}D volume into 1[+1]D (space[*time]) matrix.")
    return data.reshape(((-1,) + data.shape[3:]), order="F")


def mat_to_vol(data, shape=None, asdata=None):
    """
    Reshape nD data (normally 2D) using either shape or data shape).

    Parameters
    ----------
    data : numpy.ndarray
        Data to be reshaped.
    shape : None or list of int, optional
        New shape.
    asdata : None or numpy.ndarray, optional
        Numpy ndarray to use the shape of.

    Returns
    -------
    numpy.ndarray
        Reshaped data.

    Raises
    ------
    ValueError
        If both shape and asdata are empty.
    """
    if asdata is not None:
        if shape is not None:
            LGR.warning(
                "Both shape and asdata were defined. "
                f"Overwriting shape {shape} with asdata {asdata.shape}"
            )
        shape = asdata.shape
    elif shape is None:
        raise ValueError(
            "Both shape and asdata are empty. " "Must specify at least one"
        )

    LGR.info(f"Reshape {data.ndim}D matrix into volume with shape {shape}.")
    return data.reshape(shape, order="F")


def apply_mask(data, mask):
    """
    Reduce shape and size of data based on mask.

    Uses the concept of "mask" as defined in neuroimaging, not in numpy:
    every voxel that is not 0 is kept, the 0 are excluded.

    Parameters
    ----------
    data : numpy.ndarray
        Data to be masked.
    mask : numpy.ndarray
        Mask to be applied - where all elements that are 0 are eliminated.

    Returns
    -------
    numpy.ndarray
        Masked (data.ndim-1 Dimensions) array.

    Raises
    ------
    ValueError
        If the first mask.ndim dimensions of data have a different shape from mask.
    """
    if data.shape[: mask.ndim] != mask.shape:
        raise ValueError(
            f"Cannot mask data with shape {data.shape} using mask "
            f"with shape {mask.shape}"
        )
    if (data.ndim - mask.ndim) > 1:
        LGR.warning(f"Returning volume with {data.ndim-mask.ndim+1} dimensions.")
    else:
        LGR.info(f"Returning {data.ndim-mask.ndim+1}D array.")

    mask = mask != 0
    return data[mask]


def unmask(data, mask, shape=None, asdata=None):
    """
    Unmask 1D or 2D into an nD based on shape or asdata.

    Parameters
    ----------
    data : numpy.ndarray (1D or 2D)
        The data to unmask.
    mask : numpy.ndarray
        The nD mask to use to unwrap (unmask) the data.
        All elements different from 0 will be used as entries for data.
    shape : None or list of int, optional
        List indicating shape of nD array.
    asdata : None or numpy.ndarray, optional
        Array to take the same shape of.

    Returns
    -------
    numpy.ndarray
        The unmasked nD array version of data.

    Raises
    ------
    ValueError
        If both `shape` and `asdata` are empty
        If the first dimension of `data` and the number of available voxels in
        mask do not match.
        If the mask shape does not match the first (mask)
    """
    if asdata is not None:
        if shape is not None:
            LGR.warning(
                "Both shape and asdata were defined. "
                f"Overwriting shape {shape} with asdata {asdata.shape}"
            )
        shape = asdata.shape
    elif shape is None:
        raise ValueError(
            "Both shape and asdata are empty. " "Must specify at least one."
        )

    if shape[: mask.ndim] != mask.shape:
        raise ValueError(
            f"Cannot unmask data into shape {shape} using mask "
            f"with shape {mask.shape}"
        )
    if data.ndim > 1 and (data.shape[0] != mask.sum()):
        raise ValueError(
            "Cannot unmask data with first dimension "
            f"{data.shape[0]} using mask with "
            f"{mask.sum()} entries)"
        )

    LGR.info(f"Unmasking matrix into volume of shape {shape}")
    mask = mask != 0
    out = np.zeros(shape, dtype="float32")
    out[mask] = data
    return out


def apply_atlas(data, atlas, mask=None):
    """
    Extract average timeseries from an atlas.

    Parameters
    ----------
    data : numpy.ndarray
        A 3- or 4- D matrix (normally nifti data) of timeseries.
    atlas : numpy.ndarray
        A 2- or 3- D matrix representing the atlas, each parcel represented by a different int.
    mask : None or numpy.ndarray, optional
        A 2- or 3- D matrix representing a mask, all voxels == 0 are excluded from the computation.

    Returns
    -------
    numpy.ndarray
        A [data.ndim-1]D matrix representing the average timeseries of each parcels.

    Raises
    ------
    NotImplementedError
        If atlas is 4+ D
    ValueError
        If atlas or mask have a different shape than the first dimensions of data
    """
    if mask is None:
        mask = (data != 0).any(axis=-1)
    else:
        # Ensure that mask is boolean
        mask = mask != 0

    # #!# Add nilearn's fetching atlases utility

    if atlas.ndim > 3:
        raise NotImplementedError(
            f"Files with {atlas.ndim} dimensions are not " "supported as atlases."
        )
    if data.shape[: mask.ndim] != mask.shape:
        raise ValueError(
            f"Cannot mask data with shape {data.shape} using mask "
            f"with shape {mask.shape}"
        )
    if data.shape[: atlas.ndim] != atlas.shape:
        raise ValueError(
            f"Cannot apply atlas with shape {atlas.shape} on data "
            f"with shape {data.shape}"
        )
    if (data.ndim - atlas.ndim) > 1:
        LGR.warning(f"returning volume with {data.ndim-atlas.ndim+1} dimensions.")
    else:
        LGR.info(
            f"Returning {data.ndim-atlas.ndim+1}D array of signal averages "
            f"in atlas {atlas}."
        )

    # Mask data and atlas first
    atlas = atlas * mask
    labels = np.unique(atlas)
    labels = labels[labels > 0]
    LGR.info(f"Labels: {labels}, numbers: {len(labels)}")
    # Initialise dataframe and dictionary for series
    parcels = np.empty([len(labels), data.shape[-1]], dtype="float32")

    # Compute averages
    for n, l in enumerate(labels):
        parcels[n, :] = data[atlas == l].mean(axis=0)

    return parcels


def unfold_atlas(data, atlas, mask=None):
    """
    Return a lower dimensional matrix into a 3- or 4- D matrix based on an atlas.

    (i.e. unfold a matrix into a 3D atlas)

    Parameters
    ----------
    data : numpy.ndarray
        A 1- or 2- D matrix of shape parcels*timepoints.
    atlas : numpy.ndarray
        A 3D matrix that represents an atlas.
    mask : None or numpy.ndarray, optional
        A matrix with the same shape as atlas. All non-zero elements will be
        considered for the unfolding, all zero elements will be excluded.

    Returns
    -------
    numpy.ndarray
        A 3- or 4- D matrix of shape [atlas]*[timepoints]. Contains the
        timeseries unfolded in the atlas.

    Raises
    ------
    ValueError
        If atlas and mask have dimensions that are too different (i.e. more than
        1 dimension of difference)
        If mask has different shapes from atlas.
        If the first dimension of the data is not equal to the amount of label
        in the atlas.
    """
    if mask is None:
        mask = atlas != 0
    else:
        # Check that mask contains bool
        mask = mask != 0

    if (mask.ndim - atlas.ndim) == 1:
        atlas = atlas[..., np.newaxis]
    elif (atlas.ndim - mask.ndim) == 1:
        mask = mask[..., np.newaxis]
    elif abs(mask.ndim - atlas.ndim) > 1:
        raise ValueError(f"Cannot use {mask.ndim}D mask on {atlas.ndim}D atlas.")

    if atlas.shape[: mask.ndim] != mask.shape:
        raise ValueError(
            f"Cannot mask atlas with shape {atlas.shape} using mask "
            f"with shape {mask.shape}"
        )
    atlas = atlas * mask

    labels = np.unique(atlas)
    labels = labels[labels > 0]
    if data.shape[0] != len(labels):
        raise ValueError(
            f"Cannot unfold data with shape {data.shape} on atlas "
            f"with {len(labels)} parcels"
        )

    LGR.info(f"Unmasking data into atlas-like volume of {3+data.ndim-1} dimensions.")
    out = np.zeros_like(atlas, dtype="float32")

    for ax in range(1, data.ndim):
        if data.shape[ax] > 1:
            out = out[..., np.newaxis].repeat(data.shape[ax], axis=-1)

    for n, l in enumerate(labels):
        out[atlas == l] = data[n, ...]

    return out


"""
Copyright 2022, Stefano Moia.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
